@@ -48,6 +48,13 @@ type nodeType =
   | DOCUMENT_FRAGMENT
   | NOTATION
 
+val document_position_disconnected : int
+val document_position_preceding : int
+val document_position_following : int
+val document_position_contains : int
+val document_position_contained_by : int
+val document_position_implementation_specific : int
+
 (** Specification of [Node] objects. *)
 class type node = object
   method nodeName : js_string t readonly_prop
@@ -66,17 +73,7 @@ class type node = object
   method appendChild : node t -> node t meth
   method hasChildNodes : bool t meth
   method cloneNode : bool t -> node t meth
-end
-
-(** Specification of [TokenList] objects. *)
-class type tokenList = object
-  method length : int readonly_prop
-  method item : int -> js_string t optdef meth
-  method contains : js_string t -> bool t meth
-  method add : js_string t -> unit meth
-  method remove : js_string t -> unit meth
-  method toggle : js_string t -> bool t meth
-  method stringifier : js_string t prop
+  method compareDocumentPosition : node t -> int meth
 end
 
 (** Specification of [Attr] objects. *)
@@ -89,11 +86,11 @@ class type attr = object
 end
 
 (** Specification of [NamedNodeMap] objects. *)
-and namedNodeMap = object
-  method getNamedItem : js_string t -> node t opt meth
-  method setNamedItem : node t -> node t opt meth
-  method removeNamedItem : js_string t -> node t opt meth
-  method item : int -> node t opt meth
+and ['node] namedNodeMap = object
+  method getNamedItem : js_string t -> 'node t opt meth
+  method setNamedItem : 'node t -> 'node t opt meth
+  method removeNamedItem : js_string t -> 'node t opt meth
+  method item : int -> 'node t opt meth
   method length : int readonly_prop
 end
 
@@ -106,8 +103,7 @@ and element = object
   method removeAttribute : js_string t -> unit meth
   method hasAttribute : js_string t -> bool t meth
   method getElementsByTagName : js_string t -> element nodeList t meth
-  method classList : tokenList t readonly_prop
-  method attributes : namedNodeMap t readonly_prop
+  method attributes : attr namedNodeMap t readonly_prop
 end
 
 (** Specification of [CharacterData] objects. *)
@@ -119,7 +115,7 @@ class type characterData = object
   method appendData : js_string t -> unit meth
   method insertData : int -> js_string t -> unit meth
   method deleteData : int -> int -> unit meth
-  method replaceData : int -> int -> js_string t meth
+  method replaceData : int -> int -> js_string t -> unit meth
 end
 
 (** Specification of [Text] objects. *)
@@ -179,3 +175,53 @@ module CoerceTo : sig
   val text : #node t -> text t opt
   val attr : #node t -> attr t opt
 end
+
+(** {2 Events} *)
+
+type (-'a, -'b) event_listener
+  (** The type of event listener functions.  The first type parameter
+      ['a] is the type of the target object; the second parameter
+      ['b] is the type of the event object. *)
+
+class type ['a] event = object
+  method _type : js_string t readonly_prop
+  method target : 'a t optdef readonly_prop
+  method currentTarget : 'a t optdef readonly_prop
+
+  (* Legacy methods *)
+  method srcElement : 'a t optdef readonly_prop
+end
+
+(** {2 Event handlers} *)
+
+val no_handler : ('a, 'b) event_listener
+  (** Void event handler (Javascript [null] value). *)
+val handler : (('e #event t as 'b) -> bool t) -> ('a, 'b) event_listener
+  (** Create an event handler that invokes the provided function.
+      If the handler returns false, the default action is prevented. *)
+val full_handler : ('a -> ('e #event t as 'b) -> bool t) -> ('a, 'b) event_listener
+  (** Create an event handler that invokes the provided function.
+      The event target (implicit parameter [this]) is also passed as
+      argument to the function.  *)
+val invoke_handler : ('a, 'b) event_listener -> 'a -> 'b -> bool t
+  (** Invoke an existing handler.  Useful to chain event handlers. *)
+
+val eventTarget : (< .. > as 'a) #event t -> 'a t
+  (** Returns which object is the target of this event. *)
+
+type event_listener_id
+
+module Event : sig
+  type 'a typ
+  val make : string -> 'a typ
+end
+
+val addEventListener :
+  (< .. > t as 'a) -> 'b Event.typ ->
+  ('a, 'b) event_listener -> bool t -> event_listener_id
+  (** Add an event listener.  This function matches the
+      [addEventListener] DOM method, except that it returns
+      an id for removing the listener. *)
+
+val removeEventListener : event_listener_id -> unit
+  (** Remove the given event listener. *)
