@@ -63,6 +63,7 @@ class type cssStyleDeclaration = object
   method content : js_string t prop
   method counterIncrement : js_string t prop
   method counterReset : js_string t prop
+  method cssFloat : js_string t prop
   method cssText : js_string t prop
   method cursor : js_string t prop
   method direction : js_string t prop
@@ -123,7 +124,7 @@ class type cssStyleDeclaration = object
   method zIndex : js_string t prop
 end
 
-type ('a, 'b) event_listener = ('a, 'b optdef -> bool t) meth_callback opt
+type ('a, 'b) event_listener = ('a, 'b) Dom.event_listener
 
 type mouse_button =
   | No_button
@@ -132,9 +133,7 @@ type mouse_button =
   | Right_button
 
 class type event = object
-  method _type : js_string t readonly_prop
-  method target : element t optdef readonly_prop
-  method srcElement : element t optdef readonly_prop
+  inherit [element] Dom.event
 end
 
 and mouseEvent = object
@@ -162,6 +161,10 @@ and keyboardEvent = object
   method charCode : int optdef readonly_prop
   method keyCode : int readonly_prop
   method keyIdentifier : js_string t optdef readonly_prop
+  method altKey : bool t readonly_prop
+  method shiftKey : bool t readonly_prop
+  method ctrlKey : bool t readonly_prop
+  method metaKey : bool t readonly_prop
 end
 
 and wheelEvent = object (* All browsers but Firefox *)
@@ -179,6 +182,34 @@ and mouseScrollEvent = object (* Firefox *)
   method _VERTICAL_AXIS : int optdef readonly_prop
 end
 
+and touchEvent = object
+  inherit event
+  method touches : touchList t readonly_prop
+  method targetTouches : touchList t readonly_prop
+  method changedTouches : touchList t readonly_prop
+  method ctrlKey : bool t readonly_prop
+  method shiftKey : bool t readonly_prop
+  method altKey : bool t readonly_prop
+  method metaKey : bool t readonly_prop
+  method relatedTarget : element t opt optdef readonly_prop
+end
+
+and touchList = object
+  method length : int readonly_prop
+  method item : int -> touch t optdef meth
+end
+
+and touch = object
+  method identifier : int readonly_prop
+  method target : element t optdef readonly_prop
+  method screenX : int readonly_prop
+  method screenY : int readonly_prop
+  method clientX : int readonly_prop
+  method clientY : int readonly_prop
+  method pageX : int readonly_prop
+  method pageY : int readonly_prop
+end
+
 and eventTarget = object ('self)
   method onclick : ('self t, mouseEvent t) event_listener writeonly_prop
   method ondblclick : ('self t, mouseEvent t) event_listener writeonly_prop
@@ -200,28 +231,45 @@ end
 and storageEvent = object
   inherit event
   method key : js_string t readonly_prop
-  method oldValue : js_string t optdef readonly_prop
-  method keynewValue : js_string t optdef readonly_prop
+  method oldValue : js_string t opt readonly_prop
+  method keynewValue : js_string t opt readonly_prop
   method url : js_string t readonly_prop
-  method storageArea : storage t optdef readonly_prop
+  method storageArea : storage t opt readonly_prop
 end
 
 and storage = object
   method length : int readonly_prop
-  method key : int -> js_string t optdef meth
-  method getItem : js_string t -> js_string t optdef meth
-  method setItem : js_string t -> unit meth
+  method key : int -> js_string t opt meth
+  method getItem : js_string t -> js_string t opt meth
+  method setItem : js_string t -> js_string t -> unit meth
   method removeItem : js_string t -> unit meth
   method clear : unit meth
 end
 
+and nodeSelector = object
+  method querySelector : js_string t -> element t opt meth
+  method querySelectorAll : js_string t -> element Dom.nodeList t meth
+end
+
+and tokenList = object
+  method length : int readonly_prop
+  method item : int -> js_string t optdef meth
+  method contains : js_string t -> bool t meth
+  method add : js_string t -> unit meth
+  method remove : js_string t -> unit meth
+  method toggle : js_string t -> bool t meth
+  method stringifier : js_string t prop
+end
+
 and element = object
   inherit Dom.element
+  inherit nodeSelector
   method id : js_string t prop
   method title : js_string t prop
   method lang : js_string t prop
   method dir : js_string t prop
   method className : js_string t prop
+  method classList : tokenList t readonly_prop
   method style : cssStyleDeclaration t prop
 
   method innerHTML : js_string t prop
@@ -262,70 +310,38 @@ and clientRectList = object
   method item : int -> clientRect t optdef meth
 end
 
-let no_handler : ('a, 'b) event_listener = Js.null
-let window_event () : #event t = Js.Unsafe.variable "event"
-(* The function preventDefault must be called explicitely when
-   using addEventListener... *)
-let handler f =
-  Js.some (Js.wrap_callback
-    (fun e ->
-       Optdef.case e
-         (fun () ->
-            let e = window_event () in
-            let res = f e in
-            e##returnValue <- res; res)
-         (fun e ->
-            let res = f e in
-            if not (Js.to_bool res) then
-              (Js.Unsafe.coerce e)##preventDefault ();
-            res)))
-let full_handler f =
-  Js.some (Js.wrap_meth_callback
-    (fun this e ->
-       Optdef.case e
-         (fun () ->
-            let e = window_event () in
-            let res = f this e in
-            e##returnValue <- res; res)
-         (fun e ->
-            let res = f this e in
-            if not (Js.to_bool res) then
-              (Js.Unsafe.coerce e)##preventDefault ();
-            res)))
-let invoke_handler
-  (f : ('a, 'b) event_listener) (this : 'a) (event : 'b) : bool t =
-  Js.Unsafe.call f this [|Js.Unsafe.inject event|]
+let no_handler : ('a, 'b) event_listener = Dom.no_handler
+let handler = Dom.handler
+let full_handler = Dom.full_handler
+let invoke_handler = Dom.invoke_handler
 
 module Event = struct
-  type 'a typ = js_string t
-  let click = Js.string "click"
-  let dblclick = Js.string "dblclick"
-  let mousedown = Js.string "mousedown"
-  let mouseup = Js.string "mouseup"
-  let mouseover = Js.string "mouseover"
-  let mousemove = Js.string "mousemove"
-  let mouseout = Js.string "mouseout"
-  let keypress = Js.string "keypress"
-  let keydown = Js.string "keydown"
-  let keyup = Js.string "keyup"
-  let mousewheel = Js.string "mousewheel"
-  let _DOMMouseScroll = Js.string "DOMMouseScroll"
+  type 'a typ = 'a Dom.Event.typ
+  let click = Dom.Event.make "click"
+  let dblclick = Dom.Event.make "dblclick"
+  let mousedown = Dom.Event.make "mousedown"
+  let mouseup = Dom.Event.make "mouseup"
+  let mouseover = Dom.Event.make "mouseover"
+  let mousemove = Dom.Event.make "mousemove"
+  let mouseout = Dom.Event.make "mouseout"
+  let keypress = Dom.Event.make "keypress"
+  let keydown = Dom.Event.make "keydown"
+  let keyup = Dom.Event.make "keyup"
+  let mousewheel = Dom.Event.make "mousewheel"
+  let _DOMMouseScroll = Dom.Event.make "DOMMouseScroll"
+  let touchstart = Dom.Event.make "touchstart"
+  let touchmove = Dom.Event.make "touchmove"
+  let touchend = Dom.Event.make "touchend"
+  let touchcancel = Dom.Event.make "touchcancel"
+
+  let make = Dom.Event.make
 end
 
-type event_listener_id = unit -> unit
+type event_listener_id = Dom.event_listener_id
 
-let addEventListener (e : #eventTarget t) typ h capt =
-  if (Js.Unsafe.coerce e)##addEventListener == Js.undefined then begin
-    let ev = (Js.string "on")##concat(typ) in
-    let callback = fun e -> Js.Unsafe.call (h, e, [||]) in
-    (Js.Unsafe.coerce e)##attachEvent(ev, callback);
-    fun () -> (Js.Unsafe.coerce e)##detachEvent(ev, callback)
-  end else begin
-    (Js.Unsafe.coerce e)##addEventListener(typ, h, capt);
-    fun () -> (Js.Unsafe.coerce e)##removeEventListener (typ, h, capt)
-  end
+let addEventListener = Dom.addEventListener
 
-let removeEventListener id = id ()
+let removeEventListener = Dom.removeEventListener
 
 class type ['node] collection = object
   method length : int readonly_prop
@@ -574,6 +590,8 @@ class type imageElement = object ('self)
   method complete : bool t prop
 
   method onload : ('self t, event t) event_listener prop
+  method onerror : ('self t, event t) event_listener prop
+  method onabort : ('self t, event t) event_listener prop
 end
 
 class type objectElement = object
@@ -687,7 +705,7 @@ class type tableElement = object
   method tHead : tableSectionElement t prop
   method tFood : tableSectionElement t prop
   method rows : tableRowElement collection t readonly_prop
-  method tbodies : tableSectionElement collection t readonly_prop
+  method tBodies : tableSectionElement collection t readonly_prop
   method align : js_string t prop
   method border : js_string t prop
   method cellPadding : js_string t prop
@@ -706,6 +724,8 @@ class type tableElement = object
   method deleteRow : int -> unit meth
 end
 
+type videoElement
+
 type context = js_string t
 let _2d_ = Js.string "2d"
 
@@ -746,9 +766,8 @@ and canvasRenderingContext2D = object
   method createPattern : imageElement t -> js_string t -> canvasPattern t meth
   method createPattern_fromCanvas :
     canvasElement t -> js_string t -> canvasPattern t meth
-(*
-  CanvasPattern createPattern(in HTMLVideoElement image, in DOMJs_String repetition);
-*)
+  method createPattern_fromVideo :
+    videoElement t -> js_string t -> canvasPattern t meth
   method lineWidth : float_prop
   method lineCap : js_string t prop
   method lineJoin : js_string t prop
@@ -806,10 +825,13 @@ and canvasRenderingContext2D = object
   method drawImage_fullFromCanvas :
     canvasElement t -> float -> float -> float -> float ->
     float -> float -> float -> float -> unit meth
-(*
-  void drawImage(in HTMLVideoElement image, in float dx, in float dy, in optional float dw, in float dh);
-  void drawImage(in HTMLVideoElement image, in float sx, in float sy, in float sw, in float sh, in float dx, in float dy, in float dw, in float dh);
-*)
+  method drawImage_fromVideoWithVideo :
+    videoElement t -> float -> float -> unit meth
+  method drawImage_fromVideoWithSize :
+    videoElement t -> float -> float -> float -> float -> unit meth
+  method drawImage_fullFromVideo :
+    videoElement t -> float -> float -> float -> float ->
+    float -> float -> float -> float -> unit meth
 
   method createImageData : int -> int -> imageData t meth
   method getImageData : float -> float -> float -> float -> imageData t meth
@@ -839,6 +861,7 @@ external pixel_set : canvasPixelArray t -> int -> int -> unit = "caml_js_set"
 
 class type document = object
   inherit [element] Dom.document
+  inherit nodeSelector
   method title : js_string t prop
   method referrer : js_string t readonly_prop
   method domain : js_string t readonly_prop
@@ -885,8 +908,8 @@ class type history = object
   method go : int opt -> unit meth
   method back : unit meth
   method forward : unit meth
-  method pushState : Js.Unsafe.any -> js_string t -> js_string t opt -> unit meth
-  method replaceState : Js.Unsafe.any -> js_string t -> js_string t opt -> unit meth
+  method pushState : 'a. 'a -> js_string t -> js_string t opt -> unit meth
+  method replaceState : 'a. 'a -> js_string t -> js_string t opt -> unit meth
 end
 
 class type undoManager = object
@@ -895,12 +918,25 @@ end
 class type selection = object
 end
 
+class type navigator = object
+  method appCodeName : js_string t readonly_prop
+  method appName : js_string t readonly_prop
+  method appVersion : js_string t readonly_prop
+  method cookieEnabled : bool t readonly_prop
+  method online : bool t readonly_prop
+  method platform : js_string t readonly_prop
+  method userAgent : js_string t readonly_prop
+  method language : js_string t optdef readonly_prop
+  method userLanguage : js_string t optdef readonly_prop
+end
+
 class type window = object
   method document : document t readonly_prop
   method name : js_string t prop
   method location : location t readonly_prop
   method history : history t readonly_prop
   method undoManager : undoManager t readonly_prop
+  method navigator : navigator t
   method getSelection : selection t meth
   method close : unit meth
   method stop : unit meth
@@ -908,13 +944,14 @@ class type window = object
   method blur : unit meth
   method scroll : int -> int -> unit meth
 
-  method sessionStorage : storage t readonly_prop
-  method localStorage : storage t readonly_prop
+  method sessionStorage : storage t optdef readonly_prop
+  method localStorage : storage t optdef readonly_prop
 
   method top : window t readonly_prop
   method parent : window t readonly_prop
   method frameElement : element t opt readonly_prop
 
+  method open_ : js_string t -> js_string t -> js_string t opt -> window meth
   method alert : js_string t -> unit meth
   method confirm : js_string t -> bool t meth
   method prompt : js_string t -> js_string t -> js_string t meth
@@ -1190,16 +1227,7 @@ end
 
 (****)
 
-let eventTarget (e : #event t) =
-  let target =
-    Optdef.get (e##target) (fun () ->
-    Optdef.get (e##srcElement) (fun () -> assert false))
-  in
-  (* Workaround for Safari bug *)
-  if target##nodeType == Dom.TEXT then
-    Js.Unsafe.coerce (Opt.get (target##parentNode) (fun () -> assert false))
-  else
-    target
+let eventTarget = Dom.eventTarget
 
 let eventRelatedTarget (e : #mouseEvent t) =
   Optdef.get (e##relatedTarget) (fun () ->
@@ -1268,6 +1296,8 @@ let addMousewheelEventListener e h capt =
               h (e :> mouseEvent t) ~dx:0 ~dy:d))
       capt
 
+let element : #Dom.element t -> element t = Js.Unsafe.coerce
+
 type taggedElement =
   | A of anchorElement t
   | Area of areaElement t
@@ -1329,67 +1359,134 @@ type taggedElement =
   | Ul of uListElement t
   | Other of element t
 
+let other e = Other (e : #element t :> element t)
+
 let tagged (e : #element t) =
-  match Js.to_string (e##tagName##toLowerCase()) with
-  | "a" -> A (Js.Unsafe.coerce e)
-  | "area" -> Area (Js.Unsafe.coerce e)
-  | "base" -> Base (Js.Unsafe.coerce e)
-  | "blockquote" -> Blockquote (Js.Unsafe.coerce e)
-  | "body" -> Body (Js.Unsafe.coerce e)
-  | "br" -> Br (Js.Unsafe.coerce e)
-  | "button" -> Button (Js.Unsafe.coerce e)
-  | "canvas" -> Canvas (Js.Unsafe.coerce e)
-  | "caption" -> Caption (Js.Unsafe.coerce e)
-  | "col" -> Col (Js.Unsafe.coerce e)
-  | "colgroup" -> Colgroup (Js.Unsafe.coerce e)
-  | "del" -> Del (Js.Unsafe.coerce e)
-  | "div" -> Div (Js.Unsafe.coerce e)
-  | "dl" -> Dl (Js.Unsafe.coerce e)
-  | "fieldset" -> Fieldset (Js.Unsafe.coerce e)
-  | "form" -> Form (Js.Unsafe.coerce e)
-  | "frameset" -> Frameset (Js.Unsafe.coerce e)
-  | "frame" -> Frame (Js.Unsafe.coerce e)
-  | "h1" -> H1 (Js.Unsafe.coerce e)
-  | "h2" -> H2 (Js.Unsafe.coerce e)
-  | "h3" -> H3 (Js.Unsafe.coerce e)
-  | "h4" -> H4 (Js.Unsafe.coerce e)
-  | "h5" -> H5 (Js.Unsafe.coerce e)
-  | "h6" -> H6 (Js.Unsafe.coerce e)
-  | "head" -> Head (Js.Unsafe.coerce e)
-  | "hr" -> Hr (Js.Unsafe.coerce e)
-  | "html" -> Html (Js.Unsafe.coerce e)
-  | "iframe" -> Iframe (Js.Unsafe.coerce e)
-  | "img" -> Img (Js.Unsafe.coerce e)
-  | "input" -> Input (Js.Unsafe.coerce e)
-  | "ins" -> Ins (Js.Unsafe.coerce e)
-  | "label" -> Label (Js.Unsafe.coerce e)
-  | "legend" -> Legend (Js.Unsafe.coerce e)
-  | "li" -> Li (Js.Unsafe.coerce e)
-  | "link" -> Link (Js.Unsafe.coerce e)
-  | "map" -> Map (Js.Unsafe.coerce e)
-  | "meta" -> Meta (Js.Unsafe.coerce e)
-  | "object" -> Object (Js.Unsafe.coerce e)
-  | "ol" -> Ol (Js.Unsafe.coerce e)
-  | "optgroup" -> Optgroup (Js.Unsafe.coerce e)
-  | "option" -> Option (Js.Unsafe.coerce e)
-  | "p" -> P (Js.Unsafe.coerce e)
-  | "param" -> Param (Js.Unsafe.coerce e)
-  | "pre" -> Pre (Js.Unsafe.coerce e)
-  | "q" -> Q (Js.Unsafe.coerce e)
-  | "script" -> Script (Js.Unsafe.coerce e)
-  | "select" -> Select (Js.Unsafe.coerce e)
-  | "style" -> Style (Js.Unsafe.coerce e)
-  | "table" -> Table (Js.Unsafe.coerce e)
-  | "tbody" -> Tbody (Js.Unsafe.coerce e)
-  | "td" -> Td (Js.Unsafe.coerce e)
-  | "textarea" -> Textarea (Js.Unsafe.coerce e)
-  | "tfoot" -> Tfoot (Js.Unsafe.coerce e)
-  | "th" -> Th (Js.Unsafe.coerce e)
-  | "thead" -> Thead (Js.Unsafe.coerce e)
-  | "title" -> Title (Js.Unsafe.coerce e)
-  | "tr" -> Tr (Js.Unsafe.coerce e)
-  | "ul" -> Ul (Js.Unsafe.coerce e)
-  | _   -> Other (e : #element t :> element t)
+  let tag = Js.to_bytestring (e##tagName##toLowerCase()) in
+  if String.length tag = 0 then
+    other e
+  else
+    match String.unsafe_get tag 0 with
+      'a' ->
+        begin match tag with
+        | "a" -> A (Js.Unsafe.coerce e)
+        | "area" -> Area (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'b' ->
+        begin match tag with
+        | "base" -> Base (Js.Unsafe.coerce e)
+        | "blockquote" -> Blockquote (Js.Unsafe.coerce e)
+        | "body" -> Body (Js.Unsafe.coerce e)
+        | "br" -> Br (Js.Unsafe.coerce e)
+        | "button" -> Button (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'c' ->
+        begin match tag with
+        | "canvas" -> Canvas (Js.Unsafe.coerce e)
+        | "caption" -> Caption (Js.Unsafe.coerce e)
+        | "col" -> Col (Js.Unsafe.coerce e)
+        | "colgroup" -> Colgroup (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'd' ->
+        begin match tag with
+        | "del" -> Del (Js.Unsafe.coerce e)
+        | "div" -> Div (Js.Unsafe.coerce e)
+        | "dl" -> Dl (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'f' ->
+        begin match tag with
+        | "fieldset" -> Fieldset (Js.Unsafe.coerce e)
+        | "form" -> Form (Js.Unsafe.coerce e)
+        | "frameset" -> Frameset (Js.Unsafe.coerce e)
+        | "frame" -> Frame (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'h' ->
+        begin match tag with
+        | "h1" -> H1 (Js.Unsafe.coerce e)
+        | "h2" -> H2 (Js.Unsafe.coerce e)
+        | "h3" -> H3 (Js.Unsafe.coerce e)
+        | "h4" -> H4 (Js.Unsafe.coerce e)
+        | "h5" -> H5 (Js.Unsafe.coerce e)
+        | "h6" -> H6 (Js.Unsafe.coerce e)
+        | "head" -> Head (Js.Unsafe.coerce e)
+        | "hr" -> Hr (Js.Unsafe.coerce e)
+        | "html" -> Html (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'i' ->
+        begin match tag with
+        | "iframe" -> Iframe (Js.Unsafe.coerce e)
+        | "img" -> Img (Js.Unsafe.coerce e)
+        | "input" -> Input (Js.Unsafe.coerce e)
+        | "ins" -> Ins (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'l' ->
+        begin match tag with
+        | "label" -> Label (Js.Unsafe.coerce e)
+        | "legend" -> Legend (Js.Unsafe.coerce e)
+        | "li" -> Li (Js.Unsafe.coerce e)
+        | "link" -> Link (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'm' ->
+        begin match tag with
+        | "map" -> Map (Js.Unsafe.coerce e)
+        | "meta" -> Meta (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'o' ->
+        begin match tag with
+        | "object" -> Object (Js.Unsafe.coerce e)
+        | "ol" -> Ol (Js.Unsafe.coerce e)
+        | "optgroup" -> Optgroup (Js.Unsafe.coerce e)
+        | "option" -> Option (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'p' ->
+        begin match tag with
+        | "p" -> P (Js.Unsafe.coerce e)
+        | "param" -> Param (Js.Unsafe.coerce e)
+        | "pre" -> Pre (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'q' ->
+        begin match tag with
+        | "q" -> Q (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 's' ->
+        begin match tag with
+        | "script" -> Script (Js.Unsafe.coerce e)
+        | "select" -> Select (Js.Unsafe.coerce e)
+        | "style" -> Style (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 't' ->
+        begin match tag with
+        | "table" -> Table (Js.Unsafe.coerce e)
+        | "tbody" -> Tbody (Js.Unsafe.coerce e)
+        | "td" -> Td (Js.Unsafe.coerce e)
+        | "textarea" -> Textarea (Js.Unsafe.coerce e)
+        | "tfoot" -> Tfoot (Js.Unsafe.coerce e)
+        | "th" -> Th (Js.Unsafe.coerce e)
+        | "thead" -> Thead (Js.Unsafe.coerce e)
+        | "title" -> Title (Js.Unsafe.coerce e)
+        | "tr" -> Tr (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | 'u' ->
+        begin match tag with
+        | "ul" -> Ul (Js.Unsafe.coerce e)
+        | _ -> other e
+        end
+    | _ ->
+        other e
 
 let opt_tagged e = Opt.case e (fun () -> None) (fun e -> Some (tagged e))
 
@@ -1422,3 +1519,27 @@ let stopPropagation ev =
     (e##stopPropagation)
     (fun () -> e##cancelBubble <- Js._true)
     (fun _ -> e##_stopPropagation())
+
+let _requestAnimationFrame : (unit -> unit) Js.callback -> unit =
+  Js.Unsafe.pure_expr
+    (fun _ ->
+       let w = Js.Unsafe.coerce window in
+       let l =
+         [w##requestAnimationFrame;
+          w##mozRequestAnimationFrame;
+          w##webkitRequestAnimationFrame;
+          w##oRequestAnimationFrame;
+          w##msRequestAnimationFrame]
+       in
+       try
+         let req = List.find (fun c -> Js.Optdef.test c) l in
+         fun callback -> Js.Unsafe.fun_call req [|Js.Unsafe.inject callback|]
+       with Not_found ->
+         let now () = Js.to_float (jsnew Js.date_now ()##getTime()) in
+         let last = ref (now ()) in
+         fun callback ->
+           let t = now () in
+           let dt = !last +. 1000. /. 60. -. t in
+           let dt = if dt < 0. then 0. else dt in
+           last := t;
+           ignore (window##setTimeout (callback, dt)))
