@@ -64,6 +64,7 @@ class type cssStyleDeclaration = object
   method content : js_string t prop
   method counterIncrement : js_string t prop
   method counterReset : js_string t prop
+  method cssFloat : js_string t prop
   method cssText : js_string t prop
   method cursor : js_string t prop
   method direction : js_string t prop
@@ -126,7 +127,7 @@ end
 
 (** {2 Events} *)
 
-type (-'a, -'b) event_listener
+type (-'a, -'b) event_listener = ('a, 'b) Dom.event_listener
   (** The type of event listener functions.  The first type parameter
       ['a] is the type of the target object; the second parameter
       ['b] is the type of the event object. *)
@@ -138,11 +139,7 @@ type mouse_button =
   | Right_button
 
 class type event = object
-  method _type : js_string t readonly_prop
-  method target : element t optdef readonly_prop
-
-  (* Legacy methods *)
-  method srcElement : element t optdef readonly_prop
+  inherit [element] Dom.event
 end
 
 and mouseEvent = object
@@ -171,6 +168,10 @@ and keyboardEvent = object
   method charCode : int optdef readonly_prop
   method keyCode : int readonly_prop
   method keyIdentifier : js_string t optdef readonly_prop
+  method altKey : bool t readonly_prop
+  method shiftKey : bool t readonly_prop
+  method ctrlKey : bool t readonly_prop
+  method metaKey : bool t readonly_prop
 end
 
 and wheelEvent = object (* All browsers but Firefox *)
@@ -188,6 +189,34 @@ and mouseScrollEvent = object (* Firefox *)
   method _VERTICAL_AXIS : int optdef readonly_prop
 end
 
+and touchEvent = object
+  inherit event
+  method touches : touchList t readonly_prop
+  method targetTouches : touchList t readonly_prop
+  method changedTouches : touchList t readonly_prop
+  method ctrlKey : bool t readonly_prop
+  method shiftKey : bool t readonly_prop
+  method altKey : bool t readonly_prop
+  method metaKey : bool t readonly_prop
+  method relatedTarget : element t opt optdef readonly_prop
+end
+
+and touchList = object
+  method length : int readonly_prop
+  method item : int -> touch t optdef meth
+end
+
+and touch = object
+  method identifier : int readonly_prop
+  method target : element t optdef readonly_prop
+  method screenX : int readonly_prop
+  method screenY : int readonly_prop
+  method clientX : int readonly_prop
+  method clientY : int readonly_prop
+  method pageX : int readonly_prop
+  method pageY : int readonly_prop
+end
+
 (** Common properties of event target objects: [onclick],
     [onkeypress], ... *)
 and eventTarget = object ('self)
@@ -211,32 +240,49 @@ end
 and storageEvent = object
   inherit event
   method key : js_string t readonly_prop
-  method oldValue : js_string t optdef readonly_prop
-  method keynewValue : js_string t optdef readonly_prop
+  method oldValue : js_string t opt readonly_prop
+  method keynewValue : js_string t opt readonly_prop
   method url : js_string t readonly_prop
-  method storageArea : storage t optdef readonly_prop
+  method storageArea : storage t opt readonly_prop
 end
 
 (** Storage *)
 and storage = object
   method length : int readonly_prop
-  method key : int -> js_string t optdef meth
-  method getItem : js_string t -> js_string t optdef meth
-  method setItem : js_string t -> unit meth
+  method key : int -> js_string t opt meth
+  method getItem : js_string t -> js_string t opt meth
+  method setItem : js_string t -> js_string t -> unit meth
   method removeItem : js_string t -> unit meth
   method clear : unit meth
 end
 
 (** {2 HTML elements} *)
 
+and nodeSelector = object
+  method querySelector : js_string t -> element t opt meth
+  method querySelectorAll : js_string t -> element Dom.nodeList t meth
+end
+
+and tokenList = object
+  method length : int readonly_prop
+  method item : int -> js_string t optdef meth
+  method contains : js_string t -> bool t meth
+  method add : js_string t -> unit meth
+  method remove : js_string t -> unit meth
+  method toggle : js_string t -> bool t meth
+  method stringifier : js_string t prop
+end
+
 (** Properties common to all HTML elements *)
 and element = object
   inherit Dom.element
+  inherit nodeSelector
   method id : js_string t prop
   method title : js_string t prop
   method lang : js_string t prop
   method dir : js_string t prop
   method className : js_string t prop
+  method classList : tokenList t readonly_prop
   method style : cssStyleDeclaration t prop
 
   method innerHTML : js_string t prop
@@ -527,6 +573,8 @@ class type imageElement = object ('self)
   method complete : bool t prop
 
   method onload : ('self t, event t) event_listener prop
+  method onerror : ('self t, event t) event_listener prop
+  method onabort : ('self t, event t) event_listener prop
 end
 
 class type objectElement = object
@@ -640,7 +688,7 @@ class type tableElement = object
   method tHead : tableSectionElement t prop
   method tFood : tableSectionElement t prop
   method rows : tableRowElement collection t readonly_prop
-  method tbodies : tableSectionElement collection t readonly_prop
+  method tBodies : tableSectionElement collection t readonly_prop
   method align : js_string t prop
   method border : js_string t prop
   method cellPadding : js_string t prop
@@ -659,6 +707,8 @@ class type tableElement = object
   method deleteRow : int -> unit meth
 end
 
+type videoElement
+
 (** {2 Canvas object} *)
 
 type context
@@ -701,9 +751,8 @@ and canvasRenderingContext2D = object
   method createPattern : imageElement t -> js_string t -> canvasPattern t meth
   method createPattern_fromCanvas :
     canvasElement t -> js_string t -> canvasPattern t meth
-(*
-  CanvasPattern createPattern(in HTMLVideoElement image, in DOMJs_String repetition);
-*)
+  method createPattern_fromVideo :
+    videoElement t -> js_string t -> canvasPattern t meth
   method lineWidth : float_prop
   method lineCap : js_string t prop
   method lineJoin : js_string t prop
@@ -761,10 +810,13 @@ and canvasRenderingContext2D = object
   method drawImage_fullFromCanvas :
     canvasElement t -> float -> float -> float -> float ->
     float -> float -> float -> float -> unit meth
-(*
-  void drawImage(in HTMLVideoElement image, in float dx, in float dy, in optional float dw, in float dh);
-  void drawImage(in HTMLVideoElement image, in float sx, in float sy, in float sw, in float sh, in float dx, in float dy, in float dw, in float dh);
-*)
+  method drawImage_fromVideoWithVideo :
+    videoElement t -> float -> float -> unit meth
+  method drawImage_fromVideoWithSize :
+    videoElement t -> float -> float -> float -> float -> unit meth
+  method drawImage_fullFromVideo :
+    videoElement t -> float -> float -> float -> float ->
+    float -> float -> float -> float -> unit meth
   (* Method createImageData not available in Opera *)
   method createImageData : int -> int -> imageData t meth
   method getImageData : float -> float -> float -> float -> imageData t meth
@@ -796,6 +848,7 @@ external pixel_set : canvasPixelArray t -> int -> int -> unit = "caml_js_set"
 
 class type document = object
   inherit [element] Dom.document
+  inherit nodeSelector
   method title : js_string t prop
   method referrer : js_string t readonly_prop
   method domain : js_string t readonly_prop
@@ -846,8 +899,8 @@ class type history = object
   method go : int opt -> unit meth
   method back : unit meth
   method forward : unit meth
-  method pushState : Js.Unsafe.any -> js_string t -> js_string t opt -> unit meth
-  method replaceState : Js.Unsafe.any -> js_string t -> js_string t opt -> unit meth
+  method pushState : 'a. 'a -> js_string t -> js_string t opt -> unit meth
+  method replaceState : 'a. 'a -> js_string t -> js_string t opt -> unit meth
 end
 
 (** Undo manager *)
@@ -860,6 +913,19 @@ class type selection = object
 (*...*)
 end
 
+(** Navigator information *)
+class type navigator = object
+  method appCodeName : js_string t readonly_prop
+  method appName : js_string t readonly_prop
+  method appVersion : js_string t readonly_prop
+  method cookieEnabled : bool t readonly_prop
+  method online : bool t readonly_prop
+  method platform : js_string t readonly_prop
+  method userAgent : js_string t readonly_prop
+  method language : js_string t optdef readonly_prop
+  method userLanguage : js_string t optdef readonly_prop
+end
+
 type interval_id
 type timeout_id
 
@@ -870,6 +936,7 @@ class type window = object
   method location : location t readonly_prop
   method history : history t readonly_prop
   method undoManager : undoManager t readonly_prop
+  method navigator : navigator t
   method getSelection : selection t meth
   method close : unit meth
   method stop : unit meth
@@ -877,13 +944,14 @@ class type window = object
   method blur : unit meth
   method scroll : int -> int -> unit meth
 
-  method sessionStorage : storage t readonly_prop
-  method localStorage : storage t readonly_prop
+  method sessionStorage : storage t optdef readonly_prop
+  method localStorage : storage t optdef readonly_prop
 
   method top : window t readonly_prop
   method parent : window t readonly_prop
   method frameElement : element t opt readonly_prop
 
+  method open_ : js_string t -> js_string t -> js_string t opt -> window meth
   method alert : js_string t -> unit meth
   method confirm : js_string t -> bool t meth
   method prompt : js_string t -> js_string t -> js_string t meth
@@ -906,6 +974,14 @@ end
 val window : window t
   (** The current window *)
 
+val _requestAnimationFrame : (unit -> unit) Js.callback -> unit
+  (** Call the appropriate [requestAnimationFrame] method variant
+      (depending on the navigator), or sleep for a short amount
+      of time when there no such method is provided. We currently
+      prefix the function name with as underscore as the interface of
+      this function is not completely standardized yet. Thus, we leave
+      the room to a function with a possibly refined type. *)
+
 (* {2 Frames } *)
 
 class type frameSetElement = object
@@ -947,25 +1023,22 @@ end
 (** {2 Event handlers} *)
 
 val no_handler : ('a, 'b) event_listener
-  (** Void event handler (Javascript [null] value). *)
+  (** see [Dom.no_handler] *)
 val handler : ((#event t as 'b) -> bool t) -> ('a, 'b) event_listener
-  (** Create an event handler that invokes the provided function.
-      If the handler returns false, the default action is prevented. *)
+  (** see [Dom.handler] *)
 val full_handler : ('a -> (#event t as 'b) -> bool t) -> ('a, 'b) event_listener
-  (** Create an event handler that invokes the provided function.
-      The event target (implicit parameter [this]) is also passed as
-      argument to the function.  *)
+  (** see [Dom.full_handler] *)
 val invoke_handler : ('a, 'b) event_listener -> 'a -> 'b -> bool t
-  (** Invoke an existing handler.  Useful to chain event handlers. *)
-
+  (** see [Dom.invoke_handler] *)
 val eventTarget : #event t -> element t
-  (** Returns which HTML element is the target of this event. *)
+  (** see [Dom.eventTarget] *)
+
 val eventRelatedTarget : #mouseEvent t -> element t opt
   (** Returns this event related target. *)
 
 (** Event types: [mousedown], [keypress], ... *)
 module Event : sig
-  type 'a typ
+  type 'a typ = 'a Dom.Event.typ
   val click : mouseEvent t typ
   val dblclick : mouseEvent t typ
   val mousedown : mouseEvent t typ
@@ -976,12 +1049,17 @@ module Event : sig
   val keypress : keyboardEvent t typ
   val keydown : keyboardEvent t typ
   val keyup : keyboardEvent t typ
-
   val mousewheel : wheelEvent t typ
   val _DOMMouseScroll : mouseScrollEvent t typ
+  val touchstart : touchEvent t typ
+  val touchmove : touchEvent t typ
+  val touchend : touchEvent t typ
+  val touchcancel : touchEvent t typ
+
+  val make : string -> 'a typ
 end
 
-type event_listener_id
+type event_listener_id = Dom.event_listener_id
 
 val addEventListener :
   (#eventTarget t as 'a) -> 'b Event.typ ->
@@ -1112,6 +1190,10 @@ val createCanvas : document t -> canvasElement t
 
 (** {2 Coercion functions} *)
 
+val element : #Dom.element t -> element t
+  (** Coercion from a general DOM element to an HTML element.
+      (Unsafe in general.) *)
+
 type taggedElement =
   | A of anchorElement t
   | Area of areaElement t
@@ -1195,6 +1277,7 @@ module CoerceTo : sig
   (** HTMLElement *)
 
   val element : #Dom.node t -> element t opt
+  (* null if it is not an element *)
 
   val a : #element t -> anchorElement t opt
   val area : #element t -> areaElement t opt
